"""contains the entrypoint for the command line interface"""

import json
from pathlib import Path

import typer
from pydantic import RootModel
from rich.console import Console
from typing_extensions import Annotated

from fundamend import AhbReader, Anwendungshandbuch, MessageImplementationGuide, MigReader

app = typer.Typer(name="xml2json", help="Convert XML(s) by BDEW to JSON(s)")
err_console = Console(stderr=True)  # https://typer.tiangolo.com/tutorial/printing/#printing-to-standard-error


def _convert_to_json_file(xml_file_path: Path) -> Path:
    """converts the given XML file to a JSON file and returns the path of the latter"""
    if not xml_file_path.is_file():
        raise ValueError(f"The given path {xml_file_path.absolute()} is not a file")
    is_ahb = "ahb" in xml_file_path.stem.lower()
    is_mig = "mig" in xml_file_path.stem.lower()
    if is_ahb and is_mig:
        raise ValueError(f"Cannot detect if {xml_file_path} is an AHB or MIG")
    root_model: RootModel[Anwendungshandbuch] | RootModel[MessageImplementationGuide]
    if is_ahb:
        ahb_model = AhbReader(xml_file_path).read()
        root_model = RootModel[Anwendungshandbuch](ahb_model)
    elif is_mig:
        mig_model = MigReader(xml_file_path).read()
        root_model = RootModel[MessageImplementationGuide](mig_model)
    else:
        raise ValueError(f"Seems like {xml_file_path} is neither an AHB nor a MIG")
    out_dict = root_model.model_dump(mode="json")
    json_file_path = xml_file_path.with_suffix(".json")
    with open(json_file_path, encoding="utf-8", mode="w") as outfile:
        json.dump(out_dict, outfile, indent=True, ensure_ascii=False)
    print(f"Successfully converted {xml_file_path} file to JSON {json_file_path}")
    return json_file_path


@app.command()
def main(
    xml_path: Annotated[
        Path,
        typer.Option(
            exists=True,
            file_okay=True,
            dir_okay=True,
            writable=True,
            readable=True,
            resolve_path=True,
        ),
    ],
) -> None:
    """
    converts the xml file from xml_in_path to a json file next to the .xml
    """
    assert xml_path.exists()  # ensured by typer
    if xml_path.is_dir():
        for _xml_path in xml_path.rglob("*.xml"):
            _convert_to_json_file(_xml_path)
    else:
        _convert_to_json_file(xml_path)


def cli() -> None:
    """entry point of the script defined in pyproject.toml"""
    typer.run(main)


if __name__ == "__main__":
    app()



"""
Base class for all models in the fundamend package.
"""

from pydantic import BaseModel, ConfigDict


class FundamendBaseModel(BaseModel):
    """
    Base class for all models in the fundamend package. Defines all models as frozen.
    """

    model_config = ConfigDict(frozen=True)



"""
the MigReader class in this module parses MIG XMLs and binds them to the data model
"""

import xml.etree.ElementTree as ET
from datetime import date, datetime
from pathlib import Path

from fundamend.models.messageimplementationguide import (
    Code,
    DataElement,
    DataElementGroup,
    MessageImplementationGuide,
    MigStatus,
    Segment,
    SegmentGroup,
)
from fundamend.reader.element_distinction import (
    _is_code,
    _is_data_element,
    _is_data_element_group,
    _is_segment,
    _is_segment_group,
    _is_uebertragungsdatei,
)
from fundamend.utils import lstrip


def _to_code(element: ET.Element) -> Code:
    assert _is_code(element)
    return Code(name=element.attrib["Name"], description=element.attrib["Description"] or None, value=element.text)


def _to_data_element(element: ET.Element) -> DataElement:
    assert _is_data_element(element)
    codes = []
    for child in element:
        if _is_code(child):
            codes.append(_to_code(child))
        else:
            raise ValueError(f"unexpected element: {child.tag}")
    return DataElement(
        id=element.tag,
        name=element.attrib["Name"],
        description=element.attrib["Description"] or None,
        status_std=MigStatus(element.attrib["Status_Std"]),
        status_specification=MigStatus(element.attrib["Status_Specification"]),
        format_std=element.attrib["Format_Std"],
        format_specification=element.attrib["Format_Specification"],
        codes=tuple(codes),
    )


def _to_data_element_group(element: ET.Element) -> DataElementGroup:
    assert _is_data_element_group(element)
    data_elements = []
    for child in element:
        if _is_data_element(child):
            data_elements.append(_to_data_element(child))
        else:
            raise ValueError(f"unexpected element: {child.tag}")
    return DataElementGroup(
        id=element.tag,
        name=element.attrib["Name"],
        description=element.attrib["Description"] or None,
        status_std=MigStatus(element.attrib["Status_Std"]),
        status_specification=MigStatus(element.attrib["Status_Specification"]),
        data_elements=tuple(data_elements),
    )


def _to_segment(element: ET.Element) -> Segment:
    assert _is_segment(element)
    data_elements: list[DataElement | DataElementGroup] = []
    for child in element:
        if _is_data_element_group(child):
            data_elements.append(_to_data_element_group(child))
        elif _is_data_element(child):
            data_elements.append(_to_data_element(child))
        else:
            raise ValueError(f"unexpected element: {child.tag}")
    return Segment(
        id=lstrip("S_", element.tag),
        name=element.attrib["Name"],
        description=element.attrib["Description"] or None,
        counter=element.attrib["Counter"],
        level=int(element.attrib["Level"]),
        max_rep_std=int(element.attrib["MaxRep_Std"]),
        max_rep_specification=int(element.attrib["MaxRep_Specification"]),
        status_std=MigStatus(element.attrib["Status_Std"]),
        status_specification=MigStatus(element.attrib["Status_Specification"]),
        example=element.attrib["Example"] or None,
        number=element.attrib["Number"],
        data_elements=tuple(data_elements),
    )


def _to_segment_group(element: ET.Element) -> SegmentGroup:
    assert _is_segment_group(element)
    segments_and_groups: list[SegmentGroup | Segment] = []
    for child in element:
        if _is_segment_group(child):
            segments_and_groups.append(_to_segment_group(child))
        elif _is_segment(child):
            segments_and_groups.append(_to_segment(child))
        else:
            raise ValueError(f"unexpected element: {child.tag}")
    return SegmentGroup(
        id=lstrip("G_SG", element.tag),
        name=element.attrib["Name"],
        status_std=MigStatus(element.attrib["Status_Std"]),
        status_specification=MigStatus(element.attrib["Status_Specification"]),
        counter=element.attrib["Counter"],
        level=int(element.attrib["Level"]),
        max_rep_std=int(element.attrib["MaxRep_Std"]),
        max_rep_specification=int(element.attrib["MaxRep_Specification"]),
        elements=tuple(segments_and_groups),
    )


def _get_first_tag_starting_with_m(element: ET.Element) -> ET.Element:
    for elem in element.iter():
        if elem.tag.startswith("M_"):
            return elem
    raise ValueError("No element starting with M_ found")


class MigReader:
    """
    Accesses information from an XML based message implementation guide
    """

    def __init__(self, xml_path: Path):
        """
        initialize by providing the path to the XML file
        """
        self._xml_path = xml_path
        self._element_tree = ET.parse(self._xml_path)

    def get_publishing_date(self) -> date:
        """
        returns the publishing date of the message implementation guide
        """
        root = self._element_tree.getroot()  # might be either <M_FORMAT> or <Uebertragungsdatei>
        raw_value = root.attrib["Veroeffentlichungsdatum"]  # e.g. '24.10.2023'
        result = datetime.strptime(raw_value, "%d.%m.%Y").date()
        return result

    def get_author(self) -> str:
        """
        returns the author of the message implementation guide
        """
        root = self._element_tree.getroot()  # might be either <M_FORMAT> or <Uebertragungsdatei>
        return root.attrib["Author"]

    def get_version(self) -> str:
        """
        returns the version of the message implementation guide
        """
        root = self._element_tree.getroot()  # might be either <M_FORMAT> or <Uebertragungsdatei>
        return root.attrib["Versionsnummer"]

    def get_format(self) -> str:
        """returns the format of the message implementation guide, e.g. 'UTILTS'"""
        root = self._element_tree.getroot()
        if _is_uebertragungsdatei(root):
            root = _get_first_tag_starting_with_m(root)
        return lstrip("M_", root.tag)  # converts 'M_UTILTS' to 'UTILTS'

    def _iter_segments_and_segment_groups(self, element: ET.Element) -> list[SegmentGroup | Segment]:
        """recursive function that builds a list of all segments and segment groups"""
        result: list[Segment | SegmentGroup] = []
        if _is_segment_group(element):
            result.append(_to_segment_group(element))
        elif _is_segment(element):
            result.append(_to_segment(element))
        return result

    def read(self) -> MessageImplementationGuide:
        """
        read the entire file and convert it to a MessageImplementationGuid instance
        """
        segments_and_groups = []
        root = self._element_tree.getroot()
        if _is_uebertragungsdatei(root):
            for elem in root.iter():
                if elem.tag.startswith("M_"):
                    root = elem
                    break
        for element in root:
            segments_and_groups.extend(self._iter_segments_and_segment_groups(element))

        result = MessageImplementationGuide(
            veroeffentlichungsdatum=self.get_publishing_date(),
            autor=self.get_author(),
            versionsnummer=self.get_version(),
            format=self.get_format(),
            elements=tuple(segments_and_groups),
        )
        return result



"""
Contains some utility functions that are used in the project.
"""


def lstrip(prefix: str, text: str) -> str:
    """Strip the given prefix from the given text. If the text does not start with the prefix, return the text as is.

    Args:
        prefix: The prefix to strip.
        text: The text to strip the prefix from.

    Returns:
        The text with the prefix stripped.
    """
    if text.startswith(prefix):
        return text[len(prefix) :]
    return text


def rstrip(text: str, suffix: str) -> str:
    """Strip the given suffix from the given text. If the text does not end with the suffix, return the text as is.

    Args:
        text: The text to strip the suffix from.
        suffix: The suffix to strip.

    Returns:
        The text with the suffix stripped.
    """
    if text.endswith(suffix):
        return text[: -len(suffix)]
    return text


def strip(prefix: str, text: str, suffix: str) -> str:
    """Strip the given prefix and suffix from the given text. If the text does not start with the prefix or does not
    end with the suffix, return the text as is.

    Args:
        prefix: The prefix to strip.
        text: The text to strip the prefix from.
        suffix: The suffix to strip.

    Returns:
        The text with the prefix and suffix stripped.
    """
    return lstrip(prefix, rstrip(text, suffix))



"""
helper module to create a "materialized view" (in sqlite this means: create and populate a plain table)
"""

import logging
import tempfile
from datetime import date
from itertools import groupby
from pathlib import Path
from typing import Iterable, Literal, Optional

import sqlalchemy
from efoli import get_edifact_format_version
from pydantic import BaseModel
from sqlalchemy.sql.functions import func
from sqlmodel import Session, SQLModel, create_engine, select

from fundamend import AhbReader
from fundamend import Anwendungshandbuch as PydanticAnwendungshandbuch
from fundamend.sqlmodels.anwendungshandbuch import (
    AhbHierarchyMaterialized,
    Anwendungsfall,
)
from fundamend.sqlmodels.anwendungshandbuch import Anwendungshandbuch as SqlAnwendungshandbuch
from fundamend.sqlmodels.anwendungshandbuch import (
    Code,
    DataElement,
    DataElementGroup,
    Segment,
    SegmentGroup,
    SegmentGroupLink,
)

_logger = logging.getLogger(__name__)


def create_ahb_view(session: Session) -> None:
    """
    Create a materialized view for the Anwendungshandbücher using a SQLAlchemy session.
    Warning: This is only tested for SQLite!
    """
    path_to_sql_command = Path(__file__).parent / "materialize_ahb_view.sql"

    with open(path_to_sql_command, "r", encoding="utf-8") as sql_file:
        bare_sql = sql_file.read()

    bare_statements = bare_sql.split(";")

    for bare_statement in bare_statements:
        statement = bare_statement.strip()
        if statement:
            session.execute(sqlalchemy.text(statement))
    session.commit()
    number_of_inserted_rows = session.scalar(
        select(func.count(AhbHierarchyMaterialized.id))  # type:ignore[arg-type] # pylint:disable=not-callable #
    )
    _logger.info(
        "Inserted %d rows into the materialized view %s",
        number_of_inserted_rows,
        AhbHierarchyMaterialized.__tablename__,
    )


class _PruefiValidity(BaseModel):
    """
    models how long a model, associated with a pruefidentifikator is valid
    """

    gueltig_von: Optional[date]  # inclusive start
    gueltig_bis: Optional[date]  # exclusive end
    pruefidentifikator: str

    def overlaps(self, other: "_PruefiValidity") -> bool:
        """
        returns true if the two validity periods overlap
        """
        return (
            (self.gueltig_bis is None or other.gueltig_von is None or self.gueltig_bis > other.gueltig_von)
            and (self.gueltig_von is None or other.gueltig_bis is None or self.gueltig_von < other.gueltig_bis)
            or (self.gueltig_bis is None and other.gueltig_bis is None)
            or (self.gueltig_von is None and other.gueltig_von is None)
        )


def _check_for_no_overlaps(pruefi_validities: list[_PruefiValidity]) -> None:
    """raises a value error if there are duplicates/redundancies"""
    duplicate_pruefis_for_same_gueltigkeitszeitraum = []

    for duplicate_pruefi, group in groupby(
        sorted(pruefi_validities, key=lambda x: x.pruefidentifikator), key=lambda x: x.pruefidentifikator
    ):
        group_list = list(group)
        if any(a.overlaps(b) for a, b in zip(group_list, group_list[1:])):
            duplicate_pruefis_for_same_gueltigkeitszeitraum.append(duplicate_pruefi)
    if any(duplicate_pruefis_for_same_gueltigkeitszeitraum):
        raise ValueError(
            # pylint:disable=line-too-long
            f"There are duplicate pruefidentifikators in the AHBs: {', '.join(duplicate_pruefis_for_same_gueltigkeitszeitraum)}. Dropping the source tables is not a good idea."
        )


def create_db_and_populate_with_ahb_view(
    ahb_files: Iterable[Path | tuple[Path, date, Optional[date]] | tuple[Path, Literal[None], Literal[None]]],
    drop_raw_tables: bool = False,
) -> Path:
    """
    Creates a SQLite database as temporary file, populates it with the AHBs provided and the materializes the AHB view.
    You may provide either paths to the AHB.xml files or tuples where each Path comes with a gueltig_von and gueltig_bis
    date.
    Optionally deletes the original tables to have a smaller db file (only if the prüfis are unique across all AHBs).
    Returns the path to the temporary database file.
    The calling code should move the file to a permanent location if needed.
    """
    with tempfile.NamedTemporaryFile(suffix=".sqlite", delete=False) as sqlite_file:
        sqlite_path = Path(sqlite_file.name)
    engine = create_engine(f"sqlite:///{sqlite_path}")
    SQLModel.metadata.drop_all(engine)
    SQLModel.metadata.create_all(engine)
    pruefis_added: list[_PruefiValidity] = []
    with Session(bind=engine) as session:
        for item in ahb_files:
            ahb: PydanticAnwendungshandbuch
            gueltig_von: Optional[date]
            gueltig_bis: Optional[date]
            if isinstance(item, Path):
                ahb = AhbReader(item).read()
                gueltig_von = None
                gueltig_bis = None
            elif isinstance(item, tuple):
                ahb = AhbReader(item[0]).read()
                gueltig_von = item[1]
                gueltig_bis = item[2]
            else:
                raise ValueError(f"Invalid item type in ahb_files: {type(item)}")
            sql_ahb = SqlAnwendungshandbuch.from_model(ahb)
            sql_ahb.gueltig_von = gueltig_von
            sql_ahb.gueltig_bis = gueltig_bis
            if sql_ahb.gueltig_von is not None:
                sql_ahb.edifact_format_version = get_edifact_format_version(sql_ahb.gueltig_von)
            session.add(sql_ahb)
            pruefis_added += [
                _PruefiValidity(
                    pruefidentifikator=af.pruefidentifikator, gueltig_bis=gueltig_bis, gueltig_von=gueltig_von
                )
                for af in sql_ahb.anwendungsfaelle
            ]
        session.commit()
        session.flush()
        create_ahb_view(session)
        if drop_raw_tables:
            _check_for_no_overlaps(pruefis_added)
            for model_class in [
                SqlAnwendungshandbuch,
                Anwendungsfall,
                Code,
                DataElement,
                DataElementGroup,
                Segment,
                SegmentGroup,
                SegmentGroupLink,
            ]:
                session.execute(sqlalchemy.text(f"DROP TABLE IF EXISTS {model_class.__tablename__};"))
                _logger.debug("Dropped %s", model_class.__tablename__)
        session.commit()
        session.flush()
    return sqlite_path


__all__ = ["create_db_and_populate_with_ahb_view", "create_ahb_view"]



"""
models that work together with SQLModel+SQLAlchemy
you need to install fundamend[sqlmodels] to use this sub-package
"""

# the models here do NOT inherit from the original models, because I didn't manage to fix the issue that arise:
# <frozen abc>:106: in __new__
#     ???
# E   TypeError: Anwendungshandbuch.__init_subclass__() takes no keyword arguments
#
# or
#
# ..\.tox\dev\Lib\site-packages\sqlmodel\main.py:697: in get_sqlalchemy_type
#     raise ValueError(f"{type_} has no matching SQLAlchemy type")
# E   ValueError: <class 'fundamend.models.anwendungshandbuch.Anwendungshandbuch'> has no matching SQLAlchemy type
# => you need to keep the models in sync manually by now
